//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
extern "C" {
#include <ggl.h>
}

#include "memblock.h"
#include "gstring.h"
#include "gui.h"
#include "resources.h"


extern unsigned int gui_chgcolorfilter(unsigned int dest,unsigned int src,int param);


// CLASS EDITBOX = MAIN KEYBOARD INPUT CONTROL

void gEditBox::Update(gUpdate *u)
{

	RedrawFrame();

	// UPDATE VIEWED PORTION
	
	drawsurf.x=(u->clipx<posx)? 0:(u->clipx-posx);
	drawsurf.y=(u->clipy<posy)? 0:(u->clipy-posy);
	
	if(drawsurf.x>=vieww) return;
	if(drawsurf.y>=viewh) return;
	
	int dw,dh;
	dw=u->clipx2-posx-drawsurf.x+1;
	if(dw<=0) return;
	if(drawsurf.x+dw>=vieww) dw=vieww-drawsurf.x;
	
	dh=u->clipy2-posy-drawsurf.y+1;
	if(dh<=0) return;
	if(drawsurf.y+dh>=viewh) dh=viewh-drawsurf.y;
	
	Parent->drawsurf.x=drawsurf.x+posx;
	Parent->drawsurf.y=drawsurf.y+posy;
	
	ggl_bitblt(&(Parent->drawsurf),&drawsurf,dw,dh);
	
}

// DRAWING FUNCTIONS
void gEditBox::RedrawFrame()
{
	// DRAW NON-CLIENT ITEMS DIRECTLY INTO PARENT SURFACE
		gControl *P=Parent;
		int x,y,w,h;
		
		x=ncx;
		y=ncy;
		w=ncwidth;
		h=ncheight;
		if(StyleFlags&S_HASBORDER) {

			if(StyleFlags&S_FOCUSED) {
			P->HLine(y,x,x+w-1,GGRAY12);
			P->HLine(y+h-1,x+1,x+w-1,GGRAY5);
			P->VLine(x,y+1,y+h-1,GGRAY12);
			P->VLine(x+w-1,y+1,y+h-1,GGRAY5);

			}
			else {
				P->HLine(y,x,x+w-1,GGRAY6);
				P->HLine(y+h-1,x+1,x+w-1,GGRAY4);
				P->VLine(x,y+1,y+h-1,GGRAY6);
				P->VLine(x+w-1,y+1,y+h-1,GGRAY4);
			}
			// CLEAN 1 PIXEL SEPARATION BETWEEN BORDER AND TEXT
			P->HLine(y+1,x+1,x+w-2,BkColor);
			P->HLine(y+h-2,x+1,x+w-2,BkColor);
			P->VLine(x+1,y+2,y+h-2,BkColor);
			P->VLine(x+w-2,y+2,y+h-2,BkColor);
			
			if(y+h-2>posy+viewh) {
				// CLEAN RECTANGLE UNDER TEXT
				P->Rect(x+2,posy+viewh,x+w-2,y+h-2,BkColor);
			}

		}
		else {
			if(y+h>posy+viewh) {
				// CLEAN RECTANGLE UNDER TEXT
				P->Rect(x,posy+viewh,x+w-1,y+h-1,BkColor);
			}
		}
		// DRAW SCROLL BARS
		//if(StyleFlags&S_HSCROLLBAR) DrawHScrollBar();
		if(StyleFlags&S_VSCROLLBAR) DrawVScrollBar();

}


void gEditBox::CursorOff()
{
	int count=State>>16;
	++count;
	State+=0x10000;
	if(count!=1) return;

	
	
	// TODO: ADD CLIPPING FOR CURSOR IN CASE IT GOES OUT OF BOUNDS

	// RESTORE BACKGROUND
		drawsurf.x=Cursorx-viewx;
		drawsurf.y=Cursory-viewy;
		ggl_bitblt(&drawsurf,&cursorbk,cursorbk.width,Font.TextHeight());
		
		ClearClipRegion();
		gUpdate u;
		u.clipx=Cursorx;
		u.clipy=Cursory;
		u.clipx2=u.clipx+cursorbk.width-1;
		u.clipy2=u.clipy+Font.TextHeight();
		Invalidate(&u);

}

void gEditBox::CursorOn()
{
	// TODO: ADD CLIPPING FOR CURSOR IN CASE IT GOES OUT OF BOUNDS
	int count=State>>16;
	if(count==0) return;
	--count;
	State-=0x10000;
	if(count) return;
	
// SAVE BACKGROUND
	drawsurf.x=Cursorx-viewx;
	drawsurf.y=Cursory-viewy;
	ggl_bitblt(&cursorbk,&drawsurf,cursorbk.width,Font.TextHeight());
	
// SHOW PROPER CURSOR
	char string[2]={0,0};
	string[0]=InputMode&0xff;
	ClearClipRegion();
	DrawTextBk(Cursorx-viewx,Cursory-viewy,string,&Font,GWHITE,GGRAY7);
	
	gUpdate u;
	u.clipx=Cursorx;
	u.clipy=Cursory;
	u.clipx2=u.clipx+cursorbk.width-1;
	u.clipy2=u.clipy+Font.TextHeight();
	Invalidate(&u);
	
}

void gEditBox::StartCursorBlink()
{
	if(State&ES_CURSORBLINK) return;
	State|=ES_CURSORBLINK|ES_SHOWCURSOR;
	CursorOn();
	lastblink=tmr_ticks();
}

void gEditBox::StopCursorBlink()
{
	if(!(State&ES_CURSORBLINK)) return;
	if(State&ES_SHOWCURSOR) CursorOff();
	State&=~ES_CURSORBLINK|ES_SHOWCURSOR;
}
void gEditBox::HideCursor()
{
		CursorOff();
}
void gEditBox::UnhideCursor()
{
		CursorOn();
}
// INPUT RELATED FUNCTIONS
void gEditBox::CursorUp()
{
	char a;
	int tmppos,tmptarget;
	if((State&ES_LIMITEDIT) && (CursorLine<=EditMin)) return;
		// FIND START OF PREVIOUS LINE
		tmppos=Text.PrevLine(CursorLine);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) return;
		Cursory-=Font.TextHeight();
		Cursorx=0;
		tmptarget=Targetx;
		CursorPos=CursorLine=tmppos;
		if((State&ES_LIMITEDIT) && (CursorPos<EditMin)) {
			if(Cursorx>=Targetx) tmptarget=Cursorx+1;
		}
		while(Cursorx<tmptarget) {
			a=Text.GetChar(CursorPos);
			if(a=='\n' || a=='\r') break;
			Cursorx+=Font.CharWidth(a);
			++CursorPos;
			if(!Text.IsValidOffset(CursorPos)) break;
			if((State&ES_LIMITEDIT) && (CursorPos<EditMin)) {
				if(Cursorx>=Targetx) tmptarget=Cursorx+1;
			}
		}
		
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			int shift=viewx-Cursorx;
			viewx=Cursorx;
			ScrollRight(shift);
			SetClipRegion(0,0,shift-1,viewh-1);
			DrawAllLines();
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
			ScrollLeft(shift);
			SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
			DrawAllLines();
		}		
		if(CursorLine<ViewLine) {
			viewy-=Font.TextHeight();
			ViewLine=CursorLine;
			ClearClipRegion();
			ScrollDown1Line(Cursory);
			DrawLine(CursorLine,0,viewy);
			InvalidateVScrollBar();
		}
		
		
}

// INPUT RELATED FUNCTIONS
void gEditBox::ViewUp()
{
	
		// FIND START OF PREVIOUS LINE
				
		if(ViewLine>0) {
			ViewLine=Text.PrevLine(ViewLine);
			if(ViewLine<0) return;
				viewy-=Font.TextHeight();

			ClearClipRegion();
			ScrollDown1Line(viewy);
			DrawLine(ViewLine,0,viewy);
			InvalidateVScrollBar();
		}
		
		
}



void gEditBox::CursorDown()
{
	char a;
	int tmppos;
	tmppos=Text.NextLine(CursorPos);
	if(tmppos<0) return;	// LAST LINE
	if((State&ES_LIMITEDIT) && (tmppos>EditMax)) return;
	Cursory+=Font.TextHeight();
	Cursorx=0;
	CursorLine=CursorPos=tmppos;
		while(Cursorx<Targetx) {
			a=Text.GetChar(CursorPos);
			if(a=='\n' || a=='\r') break;
			if(a==0) break;
			if((State&ES_LIMITEDIT) && (CursorPos>=EditMax)) break;
			Cursorx+=Font.CharWidth(a);
			++CursorPos;
		}
		
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			int shift=viewx-Cursorx;
			viewx=Cursorx;
			ScrollRight(shift);
			SetClipRegion(0,0,shift-1,viewh-1);
			DrawAllLines();
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
			ScrollLeft(shift);
			SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
			DrawAllLines();
		}		
		if(Cursory+Font.TextHeight()>viewy+viewh) {
			ClearClipRegion();
			ScrollUp1Line(Cursory);
			viewy+=Font.TextHeight();
			ViewLine=Text.NextLine(ViewLine);
			DrawLine(CursorLine,0,Cursory);
			InvalidateVScrollBar();
		}

		return;
}

void gEditBox::ViewDown()
{
	int count;
	int tmppos=0,prevpos=ViewLine,cy;

	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.NextLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) return;
		prevpos=tmppos;
	}
	
	cy=viewy+DispLines*Font.TextHeight();
	
	ClearClipRegion();
	ScrollUp1Line(cy);
	viewy+=Font.TextHeight();
	ViewLine=Text.NextLine(ViewLine);
	DrawLine(tmppos,0,cy);
	InvalidateVScrollBar();

}


void gEditBox::CursorRight()
{
	if((State&ES_LIMITEDIT) && (CursorPos>=EditMax)) return;
	char a=Text.GetChar(CursorPos);
	if(a=='\n' || a=='\r') {
		// GO TO NEXT LINE
		Targetx=Cursorx=0;
		CursorPos=Text.NextLine(CursorPos);
		Cursory=Cursory+Font.TextHeight();
		CursorLine=CursorPos;
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			int shift=viewx-Cursorx;
			viewx=Cursorx;
			ScrollRight(shift);
			SetClipRegion(0,0,shift-1,viewh-1);
			DrawAllLines();
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
			ScrollLeft(shift);
			SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
			DrawAllLines();
		}		
		

		if(Cursory+Font.TextHeight()>viewy+viewh) {
			ClearClipRegion();
			ScrollUp1Line(Cursory);
			viewy+=Font.TextHeight();
			ViewLine=Text.NextLine(ViewLine);
			DrawLine(CursorLine,0,Cursory);
			InvalidateVScrollBar();
		}

		return;
	}
	if(a!=0) {
	Cursorx+=Font.CharWidth(a);
	Targetx=Cursorx;
	++CursorPos;
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		int shift=viewx-Cursorx;
		viewx=Cursorx;
		ScrollRight(shift);
		SetClipRegion(0,0,shift-1,viewh-1);
		DrawAllLines();
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
		ScrollLeft(shift);
		SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
		DrawAllLines();
	}		
	

	
	}
}

void gEditBox::ViewRight()
{
		// SCROLL TO THE RIGHT
		int shift=Font.MaxWidth;
		viewx+=shift;
		ScrollLeft(shift);
		SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
		DrawAllLines();
}		
	

void gEditBox::CursorLeft()
{
	if(CursorPos==0) return;
	if((State&ES_LIMITEDIT) && (CursorPos<=EditMin)) return;
	if(CursorPos==CursorLine)
	{
		// FIND START OF PREVIOUS LINE
		CursorPos=Text.EndOfPrevLine(CursorPos);
		CursorLine=Text.StartOfLine(CursorPos);
		Cursory-=Font.TextHeight();
		Cursorx=Text.TextWidth(CursorLine,Font);
		Targetx=Cursorx;
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			int shift=viewx-Cursorx;
			viewx=Cursorx;
			ScrollRight(shift);
			SetClipRegion(0,0,shift-1,viewh-1);
			DrawAllLines();
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
			ScrollLeft(shift);
			SetClipRegion((vieww-shift)<0? 0:vieww-shift,0,vieww-1,viewh-1);
			DrawAllLines();
		}		
		
		
		if(CursorLine<ViewLine) {
			viewy-=Font.TextHeight();
			ViewLine=CursorLine;
			ClearClipRegion();
			ScrollDown1Line(Cursory);
			DrawLine(CursorLine,0,viewy);
			InvalidateVScrollBar();
		}

		return;
	}
	--CursorPos;
	Cursorx-=Font.CharWidth(Text.GetChar(CursorPos));
	Targetx=Cursorx;
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		int shift=viewx-Cursorx;
		viewx=Cursorx;
		ScrollRight(shift);
		SetClipRegion(0,0,shift-1,viewh-1);
		DrawAllLines();
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
		ScrollLeft(shift);
		SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
		DrawAllLines();
	}		
}

void gEditBox::ViewLeft()
{
	int cx;
	if(viewx==0) return;
	cx=viewx-Font.MaxWidth;
	if(cx<0) cx=0;
		// SCROLL TO THE LEFT
		int shift=viewx-cx;
		viewx=cx;
		ScrollRight(shift);
		SetClipRegion(0,0,shift-1,viewh-1);
		DrawAllLines();
		
}


void gEditBox::PageUp()
{
	char a;
	int count;
	int tmppos=0,prevpos=CursorLine;
	
	if((State&ES_LIMITEDIT) && (CursorLine<=Text.StartOfLine(EditMin))) return;
	
	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.PrevLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) {
			tmppos=prevpos;
			break;
		}
		if((State&ES_LIMITEDIT) && (tmppos<EditMin)) { ++count; break; }
		prevpos=tmppos;
	}
	
	if(tmppos==CursorLine) return;
	
		Cursory-=Font.TextHeight()*count;
		Cursorx=0;
		CursorPos=CursorLine=tmppos;
		while(Cursorx<Targetx) {
			a=Text.GetChar(CursorPos);
			if(a=='\n' || a=='\r') break;
			Cursorx+=Font.CharWidth(a);
			++CursorPos;
			if(!Text.IsValidOffset(CursorPos)) break;
			if((State&ES_LIMITEDIT) && (CursorPos<EditMin)) {
				if(Cursorx>=Targetx) Targetx=Cursorx+1;
			}

		}
		
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			viewx=Cursorx;
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
		}		
		if(CursorLine<ViewLine) {
			viewy=Cursory;
			ViewLine=CursorLine;
		}
		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		
		
	
}

void gEditBox::ViewPageUp()
{

	int count;
	int tmppos=0,prevpos=ViewLine;
	int cy;
	
	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.PrevLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) {
			tmppos=prevpos;
			break;
		}
		prevpos=tmppos;
	}
	
	if(tmppos==ViewLine) return;
	
		cy=viewy-Font.TextHeight()*count;
		viewy=cy;
		ViewLine=tmppos;
		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
	
}



void gEditBox::PageDown()
{
	char a;
	int count;
	int tmppos=0,prevpos=CursorLine;
	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.NextLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) {
			tmppos=prevpos;
			break;
		}
		if((State&ES_LIMITEDIT) && (tmppos>EditMax)) {
			tmppos=prevpos;
			break;
		}
		prevpos=tmppos;
	}
	
	if(tmppos==CursorLine) return;
	
	Cursory+=Font.TextHeight()*count;
	Cursorx=0;
	CursorLine=CursorPos=tmppos;
		while(Cursorx<Targetx) {
			a=Text.GetChar(CursorPos);
			if(a=='\n' || a=='\r') break;
			if(a==0) break;
			if((State&ES_LIMITEDIT) && (CursorPos>=EditMax)) break;
			Cursorx+=Font.CharWidth(a);
			++CursorPos;
		}
		
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			viewx=Cursorx;
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
		}		
		while(Cursory+Font.TextHeight()>viewy+viewh) {
			viewy+=Font.TextHeight();
			ViewLine=Text.NextLine(ViewLine);
		}

		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		return;	
}


void gEditBox::ViewPageDown()
{

	int count,cy;
	int tmppos=0,prevpos=ViewLine;
	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.NextLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) return;
		prevpos=tmppos;
	}
	
	cy=viewy+Font.TextHeight()*DispLines;
	

	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.NextLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) {
			tmppos=prevpos;
			break;
		}
		prevpos=tmppos;
	}
	
	cy+=Font.TextHeight()*count;

	while(cy+Font.TextHeight()>viewy+viewh) {
			viewy+=Font.TextHeight();
			ViewLine=Text.NextLine(ViewLine);
	}

		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		return;	
}





void gEditBox::Home()
{
	if((State&ES_LIMITEDIT) && (CursorLine<EditMin)) {
		MoveCursor(EditMin);
		return;
	}
	
CursorPos=CursorLine;
Cursorx=0;
Targetx=Cursorx;
if(Cursorx<viewx) {
	// SCROLL TO THE LEFT
	int shift=viewx-Cursorx;
	viewx=Cursorx;
	ScrollRight(shift);
	SetClipRegion(0,0,shift-1,viewh-1);
	DrawAllLines();
}
if(Cursorx+cursorbk.width>=viewx+vieww) {
	// SCROLL TO THE RIGHT
	int shift=Cursorx+cursorbk.width-viewx-vieww;
	viewx+=shift;
	ScrollLeft(shift);
	SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
	DrawAllLines();
}		

}

void gEditBox::ViewHome()
{
int cx=0;	
if(cx<viewx) {
	// SCROLL TO THE LEFT
	int shift=viewx-cx;
	viewx=cx;
	ScrollRight(shift);
	SetClipRegion(0,0,shift-1,viewh-1);
	DrawAllLines();
}

}



void gEditBox::ViewEnd()
{
	int tmppos=0,prevpos=ViewLine;
	int maxw=0,tmpw;
	int count;
	
	for(count=0;count<DispLines;++count) {
		// FIND START OF PREVIOUS LINE
		tmppos=Text.NextLine(prevpos);		// NOW IT POINTS TO THE BEGINNING OF THE LINE
		if(tmppos<0) {
			tmppos=prevpos;
			break;
		}
		prevpos=tmppos;
		tmpw=Text.TextWidth(tmppos,Font);
		if(tmpw>maxw) maxw=tmpw;
		
	}
	
	tmpw=maxw-vieww;
	if(tmpw<0) tmpw=0;
	
	if(tmpw<viewx) {
		// SCROLL TO THE LEFT
		int shift=viewx-tmpw;
		viewx=tmpw;
		ScrollRight(shift);
		SetClipRegion(0,0,shift-1,viewh-1);
		DrawAllLines();
		return;
	}
	if(maxw>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=maxw-viewx-vieww;
		viewx+=shift;
		ScrollLeft(shift);
		SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
		DrawAllLines();
	}		
}


void gEditBox::End()
{
	int tmppos=Text.EndOfLine(CursorPos);
	if((State&ES_LIMITEDIT) && (tmppos>EditMax))
	{
		MoveCursor(EditMax);
		return;
	}
	CursorPos=tmppos;
	Cursorx=Text.TextWidth(CursorLine,Font);
	Targetx=Cursorx;
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		int shift=viewx-Cursorx;
		viewx=Cursorx;
		ScrollRight(shift);
		SetClipRegion(0,0,shift-1,viewh-1);
		DrawAllLines();
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
		ScrollLeft(shift);
		SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
		DrawAllLines();
	}		
}

void gEditBox::StartOfText()
{
	if((State&ES_LIMITEDIT) && (EditMin>=0)) {
		MoveCursor(EditMin);
		return;
	}
	Cursory=0;
	Targetx=Cursorx=0;
	ViewLine=CursorLine=CursorPos=Text.StartOfText();
	viewx=viewy=0;
	ClearClipRegion();
	DrawAllLines();
	InvalidateVScrollBar();
	return;	
}

void gEditBox::ViewStartOfText()
{
	ViewLine=Text.StartOfText();
	viewx=viewy=0;
	ClearClipRegion();
	DrawAllLines();
	InvalidateVScrollBar();
	return;	
}

void gEditBox::EndOfText()
{
int tmpline;
if((State&ES_LIMITEDIT) && (EditMax<Text.EndOfText())) {
	MoveCursor(EditMax);
	return;
}

	CursorPos=Text.EndOfText();
	CursorLine=Text.StartOfLine(CursorPos);
	Cursorx=Text.TextWidth(CursorLine,Font);
	Targetx=Cursorx;
		
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			viewx=Cursorx;
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
		}
		
		tmpline=ViewLine;
		Cursory=viewy;
		while(tmpline!=CursorLine) {
			tmpline=Text.NextLine(tmpline);
			Cursory+=Font.TextHeight();
			if(Cursory+Font.TextHeight()>viewy+viewh) {
				viewy+=Font.TextHeight();
				ViewLine=Text.NextLine(ViewLine);
			}

		}
		
		
		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		return;	
}

void gEditBox::ViewEndOfText()
{
int tmpline;
int cpos,cline,cx,mincx,cy;
	cpos=Text.EndOfText();
	cline=Text.StartOfLine(cpos);
	cx=Text.TextWidth(cline,Font);
	mincx=cx-vieww;
	if(mincx<0) mincx=0;
	viewx=mincx;
		
		tmpline=ViewLine;
		cy=viewy;
		while(tmpline!=cline) {
			tmpline=Text.NextLine(tmpline);
			cy+=Font.TextHeight();
			if(cy+Font.TextHeight()>viewy+viewh) {
				viewy+=Font.TextHeight();
				ViewLine=Text.NextLine(ViewLine);
			}

		}
		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		return;	
}


void gEditBox::MoveCursor(int Offset)
{
	int tmpline,a;
	if(Offset<0) return;
	if(Offset>Text.EndOfText()) return;
	CursorPos=Offset;
	CursorLine=Text.StartOfLine(CursorPos);
	tmpline=CursorLine;
	Cursorx=0;

	while(tmpline<CursorPos) {
		a=Text.GetChar(tmpline);
		Cursorx+=Font.CharWidth(a);
		++tmpline;
		if(!Text.IsValidOffset(tmpline)) break;
	}

	Targetx=Cursorx;
	
			if(Cursorx<viewx) {
				// SCROLL TO THE LEFT
				viewx=Cursorx;
			}
			if(Cursorx+cursorbk.width>=viewx+vieww) {
				// SCROLL TO THE RIGHT
				int shift=Cursorx+cursorbk.width-viewx-vieww;
				viewx+=shift;
			}
			
			tmpline=ViewLine;
			Cursory=viewy;
			
			if(CursorLine>ViewLine) {
			while(tmpline!=CursorLine) {
				tmpline=Text.NextLine(tmpline);
				Cursory+=Font.TextHeight();
				if(Cursory+Font.TextHeight()>viewy+viewh) {
					viewy+=Font.TextHeight();
					ViewLine=Text.NextLine(ViewLine);
				}

			}
			}
			else {
			while(CursorLine<ViewLine) {
				viewy-=Font.TextHeight();
				ViewLine=Text.PrevLine(ViewLine);
			}
			Cursory=viewy;
			}
			ClearClipRegion();
			DrawAllLines();
			InvalidateVScrollBar();
			return;	
}


void gEditBox::InsertChar(int character)
{

	int oldcp=CursorPos;
	if(State&ES_LIMITEDIT) {
		if(CursorPos<EditMin || CursorPos>EditMax) return;
		if(State&ES_OVERWRITE && CursorPos==EditMax) return;
	}
	if(State&ES_OVERWRITE) {
		int a=Text.GetChar(CursorPos);
		
		if(a!='\n' && a!='\r' && a!=0) {
			Text.Delete(CursorPos);
			if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
			if(SelStart>=0 && CursorPos<SelStart) SelStart--;
			if(State&ES_LIMITEDIT) EditMax--;

		}
		
	}
	Text.InsertChar(CursorPos,character);
	if(SelEnd>=0 && CursorPos<SelEnd) SelEnd++;
	if(SelStart>=0 && CursorPos<SelStart) SelStart++;
	if(State&ES_LIMITEDIT) EditMax++;
	
	CursorOff();
	int oldcursorx=Cursorx,oldcursory=Cursory,oldviewx=viewx;
	
	if(character=='\n' || character=='\r') {
		// GO TO NEXT LINE
		++TotalLines;
		Targetx=Cursorx=0;
		CursorPos=Text.NextLine(CursorPos);
		Cursory=Cursory+Font.TextHeight();
		CursorLine=CursorPos;
		ClearClipRegion();
		DrawLine(oldcp,oldcursorx,oldcursory);
		if(Cursorx<viewx) {
			// SCROLL TO THE LEFT
			viewx=Cursorx;
		}
		if(Cursorx+cursorbk.width>=viewx+vieww) {
			// SCROLL TO THE RIGHT
			int shift=Cursorx+cursorbk.width-viewx-vieww;
			viewx+=shift;
		}		
		

		if(Cursory+Font.TextHeight()>viewy+viewh) {
			if(viewx==oldviewx) {
			ClearClipRegion();
			ScrollUp1Line(Cursory);
			viewy+=Font.TextHeight();
			ViewLine=Text.NextLine(ViewLine);
			DrawLine(CursorLine,0,Cursory);
			InvalidateVScrollBar();
			}
			else {
				viewy+=Font.TextHeight();
				ViewLine=Text.NextLine(ViewLine);
				InvalidateVScrollBar();
			}
		}
		else {
			if(viewx==oldviewx) {
				// UPDATE REST OF THE SCREEN
				ScrollDown1Line(Cursory);
				DrawLine(CursorPos,Cursorx,Cursory);
			}
		}

		if(viewx!=oldviewx) {
			ClearClipRegion();
			DrawAllLines();
		}
	}
	else {
	
	if(character!=0) {
	Cursorx+=Font.CharWidth(character);
	Targetx=Cursorx;
	++CursorPos;
	ClearClipRegion();
	DrawLine(oldcp,oldcursorx,oldcursory);
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		int shift=viewx-Cursorx;
		viewx=Cursorx;
		ScrollRight(shift);
		SetClipRegion(0,0,shift-1,viewh-1);
		DrawAllLines();
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
		ScrollLeft(shift);
		SetClipRegion(vieww-shift,0,vieww-1,viewh-1);
		DrawAllLines();
	}		
	

	
	}

	}
	CursorOn();
}

void gEditBox::DelChar(int backspace)
{

	char a;
	
	if(backspace) {
		
		if(CursorPos==0) return;
		if((State&ES_LIMITEDIT) && (CursorPos<=EditMin)) return;
		--CursorPos;
		a=Text.GetChar(CursorPos);
			if(a=='\n' || a=='\r') {
				
			CursorLine=Text.PrevLine(CursorPos+1);
			Cursorx=Text.TextWidth(CursorLine,Font);
			Text.Delete(CursorPos);
			if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
			if(SelStart>=0 && CursorPos<SelStart) SelStart--;
			if(SelEnd>=0 && SelStart>=0 && SelEnd<=SelStart) { SelEnd=SelStart=-1; }
			if(State&ES_LIMITEDIT) --EditMax;

			Cursory-=Font.TextHeight();
			--TotalLines;
				
			if(a=='\n') {
				// DELETE PREVIOUS '\r' IF EXISTS
				if(CursorPos==0) return;
				a=Text.GetChar(CursorPos-1);
				if(a=='\r') { Text.Delete(--CursorPos);
				if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
				if(SelStart>=0 && CursorPos<SelStart) SelStart--;
				if(SelEnd>=0 && SelStart>=0 && SelEnd<=SelStart) { SelEnd=SelStart=-1; }
				if(State&ES_LIMITEDIT) --EditMax;

				}
			}
			
			// REDRAW ALL LINES BELOW
			if(Cursorx<viewx) {
				// SCROLL TO THE LEFT
				viewx=Cursorx;
			}
			if(Cursorx+cursorbk.width>=viewx+vieww) {
				// SCROLL TO THE RIGHT
				int shift=Cursorx+cursorbk.width-viewx-vieww;
				viewx+=shift;
			}		
			if(CursorLine<ViewLine) {
				viewy=Cursory;
				ViewLine=CursorLine;
			}
			ClearClipRegion();
			DrawAllLines();
			InvalidateVScrollBar();
			CursorOn();
			return;
			
			}
			else {
				Cursorx-=Font.CharWidth(a);
				Text.Delete(CursorPos);
				if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
				if(SelStart>=0 && CursorPos<SelStart) SelStart--;
				if(SelEnd>=0 && SelStart>=0 && SelEnd<=SelStart) { SelEnd=SelStart=-1; }
				if(State&ES_LIMITEDIT) --EditMax;
				CursorOff();
				ClearClipRegion();
				DrawLine(CursorPos,Cursorx,Cursory);

				if(Cursorx<viewx) {
					// SCROLL TO THE LEFT
					int shift=viewx-Cursorx;
					viewx=Cursorx;
					ScrollRight(shift);
					SetClipRegion(0,0,shift-1,viewh-1);
					DrawAllLines();
				}
				if(Cursorx+cursorbk.width>=viewx+vieww) {
					// SCROLL TO THE RIGHT
					int shift=Cursorx+cursorbk.width-viewx-vieww;
					viewx+=shift;
					ScrollLeft(shift);
					SetClipRegion((vieww-shift)<0? 0:vieww-shift,0,vieww-1,viewh-1);
					DrawAllLines();
				}
				
				CursorOn();
				return;
				
			}
		
				
			
	}
	
	if(CursorPos>=Text.TotalUsed-1) return;
	if((State&ES_LIMITEDIT) && CursorPos>=EditMax) return;
	a=Text.GetChar(CursorPos);
		if(a=='\n' || a=='\r') {
		Text.Delete(CursorPos);
		if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
		if(SelStart>=0 && CursorPos<SelStart) SelStart--;
		if(SelEnd>=0 && SelStart>=0 && SelEnd<=SelStart) { SelEnd=SelStart=-1; }
		if(State&ES_LIMITEDIT) --EditMax;

		if((a=='\r') && (Text.GetChar(CursorPos)=='\n')) {
			Text.Delete(CursorPos);
			if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
			if(SelStart>=0 && CursorPos<SelStart) SelStart--;
			if(SelEnd>=0 && SelStart>=0 && SelEnd<=SelStart) { SelEnd=SelStart=-1; }
			if(State&ES_LIMITEDIT) --EditMax;

		}
		--TotalLines;
		// REDRAW ALL LINES
		CursorOff();
		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		CursorOn();
		return;
		}
		else {
			Text.Delete(CursorPos);
			if(SelEnd>=0 && CursorPos<SelEnd) SelEnd--;
			if(SelStart>=0 && CursorPos<SelStart) SelStart--;
			if(SelEnd>=0 && SelStart>=0 && SelEnd<=SelStart) { SelEnd=SelStart=-1; }
			if(State&ES_LIMITEDIT) --EditMax;

			CursorOff();
			DrawLine(CursorPos,Cursorx,Cursory);
			CursorOn();
			return;
		}
	

}

void gEditBox::DrawLine(int LineOffset,int _x,int _y)
{
	//ClearClipRegion();
	
	int x=_x-viewx;
	int y=_y-viewy;
	char a;
	
	

	if(y>clipy2) return;
	if(y+Font.TextHeight()<=clipy) return;

	int w,h;
	gglsurface srf;
	srf.addr=(int *)Font.FontData->Bitmap;
	srf.width=Font.FontData->BitmapWidth;
	srf.y=0;

	h=Font.TextHeight();
	
	if(y<clipy) {
		h-=clipy-y;
		srf.y=clipy-y;
		y=clipy;
	}
	if(y+h-1>clipy2) h=clipy2-y+1;

	drawsurf.y=y;
	drawsurf.x=x;
	
	while((a=Text.GetChar(LineOffset))) {
		if(a=='\n' || a=='\r' || a==0) break;
	w=Font.FontData->WidthMap[(int)(a)];
	srf.x=w&0xffff;
	w>>=16;
	if(drawsurf.x>clipx2) break;
	if(drawsurf.x+w-1<clipx) {
		drawsurf.x+=w;
		++LineOffset;
		continue;
	}
	if(drawsurf.x<clipx) {
			srf.x+=clipx-drawsurf.x;
			w-=clipx-drawsurf.x;
			drawsurf.x=clipx;
	}
	if(drawsurf.x+w-1>clipx2) w=clipx2-drawsurf.x+1;
	
	ggl_rect(&drawsurf,drawsurf.x,drawsurf.y,drawsurf.x+w-1,drawsurf.y+h-1,ggl_mkcolor( (((SelStart>=0) && (SelEnd>=0) && (LineOffset>=SelStart) && (LineOffset<SelEnd))? SelColor:BkColor)&0xf));
	if((Color&0xf)==0xf) ggl_bitbltmask(&drawsurf,&srf,w,h,0);
	else ggl_bitbltoper(&drawsurf,&srf,w,h,ggl_mkcolor(Color&0xf),&gui_chgcolorfilter);
	drawsurf.x+=w;
	++LineOffset;	

	}
	
	if(drawsurf.x<=clipx2) {
		// CLEAR UNTIL END OF LINE
		ggl_rect(&drawsurf,drawsurf.x,drawsurf.y,clipx2,drawsurf.y+h-1,ggl_mkcolor(BkColor&0xf));
	}

	gUpdate u;
	u.clipx=_x;
	u.clipy=_y;
	u.clipx2=viewx+vieww-1;
	u.clipy2=u.clipy+Font.TextHeight()-1;
	Invalidate(&u);
}
void gEditBox::ClrEOL(int x,int y)
{
	Rect(x-viewx,y-viewy,vieww-1,y-viewy+Font.TextHeight()-1,GWHITE);
}

void gEditBox::ScrollDown1Line(int ytop)
{
int y=ytop-viewy;
int h=viewh-y;

if(h>0) {
	drawsurf.x=0;
	drawsurf.y=y;
	ggl_scrolldn(&drawsurf,vieww,h,Font.TextHeight());
	gUpdate u;
	u.clipx=viewx;
	u.clipy=y+viewy;
	u.clipx2=viewx+vieww-1;
	u.clipy2=viewy+viewh-1;
	Invalidate(&u);
}
return;	
}
void gEditBox::ScrollUp1Line(int ybot)
{
	
	int y=ybot-viewy;
	if(y>viewh) y=viewh;

	if(y>0) {
		drawsurf.x=0;
		drawsurf.y=0;
		ggl_scrollup(&drawsurf,vieww,y,Font.TextHeight());
		gUpdate u;
		u.clipx=viewx;
		u.clipy=viewy;
		u.clipx2=viewx+vieww-1;
		u.clipy2=y+viewy-1;
		Invalidate(&u);
	}
	return;	
	
}

void gEditBox::ScrollLeft(int shift)
{
	if(shift<vieww) { 
		drawsurf.x=0;
		drawsurf.y=0;
		ggl_scrolllf(&drawsurf,vieww,viewh,shift);
	}
		Invalidate();
	return;	
	
}

void gEditBox::ScrollRight(int shift)
{
	if(shift<vieww) {
		drawsurf.x=0;
		drawsurf.y=0;
		ggl_scrollrt(&drawsurf,vieww,viewh,shift);
	}
		Invalidate();
	return;	
	
}


void gEditBox::DrawAllLines()
{
	//ClearClipRegion();
	
	int x=-viewx;
	int cy=0,y;
	int LineOffset=ViewLine;
	char a;
	int w,h;
	gglsurface srf;
	srf.addr=(int *)Font.FontData->Bitmap;
	srf.width=Font.FontData->BitmapWidth;
	
	do {
	y=cy;
	if(y>clipy2) break;

	h=Font.TextHeight();
	srf.y=0;
	
	if(y<clipy) {
		h-=clipy-y;
		srf.y=clipy-y;
		y=clipy;
	}
	if(y+h-1>clipy2) h=clipy2-y+1;

	drawsurf.y=y;
	drawsurf.x=x;
	
	while((a=Text.GetChar(LineOffset))) {
		if(a=='\n' || a=='\r' || a==0) break;
	w=Font.FontData->WidthMap[(int)(a)];
	srf.x=w&0xffff;
	w>>=16;
	if(drawsurf.x>clipx2) break;
	if(drawsurf.x+w-1<clipx) {
		drawsurf.x+=w;
		++LineOffset;
		continue;
	}
	if(drawsurf.x<clipx) {
			srf.x+=clipx-drawsurf.x;
			w-=clipx-drawsurf.x;
			drawsurf.x=clipx;
	}
	if(drawsurf.x+w-1>clipx2) w=clipx2-drawsurf.x+1;
	
	ggl_rect(&drawsurf,drawsurf.x,drawsurf.y,drawsurf.x+w-1,drawsurf.y+h-1,ggl_mkcolor( (((SelStart>=0) && (SelEnd>=0) && (LineOffset>=SelStart) && (LineOffset<SelEnd))? SelColor:BkColor)&0xf));
	if((Color&0xf)==0xf) ggl_bitbltmask(&drawsurf,&srf,w,h,0);
	else ggl_bitbltoper(&drawsurf,&srf,w,h,ggl_mkcolor(Color&0xf),&gui_chgcolorfilter);
	drawsurf.x+=w;
	++LineOffset;	

	}
	
	if(drawsurf.x<=clipx2) {
		// CLEAR UNTIL END OF LINE
		if(drawsurf.x<0) drawsurf.x=0;
		ggl_rect(&drawsurf,drawsurf.x,drawsurf.y,clipx2,drawsurf.y+h-1,ggl_mkcolor(BkColor&0xf));
	}

	cy+=Font.TextHeight();
	x=-viewx;
	LineOffset=Text.NextLine(LineOffset);
	} while(LineOffset>=0);
	
	if(cy<=clipy2) {
		// CLEAR UNTIL END OF CONTROL
		drawsurf.y=cy;
		drawsurf.x=0;
		ggl_rect(&drawsurf,drawsurf.x,drawsurf.y,clipx2,clipy2,ggl_mkcolor(BkColor&0xf));
	}


	Invalidate();

	
}



void gEditBox::BeginSelection()
{
SelStart=CursorPos;
if(SelEnd>=0 && SelEnd<SelStart) SelEnd=-1;
if(SelStart==SelEnd) SelStart=SelEnd=-1;
ClearClipRegion();
DrawAllLines();
}
void gEditBox::EndSelection()
{
SelEnd=CursorPos;
if(SelStart>=0 && SelStart>SelEnd) SelStart=-1;
if(SelStart==SelEnd) SelStart=SelEnd=-1;
ClearClipRegion();
DrawAllLines();
}


void gEditBox::Copy()
{
	if(SelStart<0 || SelEnd<0) return;
	if(SelEnd<=SelStart) return;
	
	gApplication *app=GetApplication();

	if(SelEnd-SelStart>app->Clipboard.TotalUsed) {
		if(!app->Clipboard.Append(SelEnd-SelStart-app->Clipboard.TotalUsed))
		{
			// NOT ENOUGH MEMORY TO USE CLIPBOARD, CLEAR CLIPBOARD AND RETURN
			app->Clipboard.Shrink(app->Clipboard.TotalUsed);
			// DO NOT ISSUE INSUF. MEMORY ERROR
			// TODO: SEND AUDIBLE SIGNAL THAT OPERATION FAILD
			return;
		}
	}
	else app->Clipboard.Shrink(app->Clipboard.TotalUsed-(SelEnd-SelStart));	// MAKE CLIPBOARD PROPER SIZE
	
	app->Clipboard.MemMove(0,Text,SelStart,SelEnd-SelStart);
	

}

void gEditBox::Cut()
{
	if(SelStart<0 || SelEnd<0) return;
	if(SelEnd<=SelStart) return;
	
	
	gApplication *app=GetApplication();

	if(SelEnd-SelStart>app->Clipboard.TotalUsed) {
		if(!app->Clipboard.Append(SelEnd-SelStart-app->Clipboard.TotalUsed))
		{
			// NOT ENOUGH MEMORY TO USE CLIPBOARD, CLEAR CLIPBOARD AND RETURN
			app->Clipboard.Shrink(app->Clipboard.TotalUsed);
			// DO NOT ISSUE INSUF. MEMORY ERROR
			// TODO: SEND AUDIBLE SIGNAL THAT OPERATION FAILED
			return;
		}
	}
	else app->Clipboard.Shrink(app->Clipboard.TotalUsed-(SelEnd-SelStart));	// MAKE CLIPBOARD PROPER SIZE
	
	app->Clipboard.MemMove(0,Text,SelStart,SelEnd-SelStart);

	
	RemoveText(SelStart,SelEnd);
	
	SelStart=SelEnd=-1;
	
	ClearClipRegion();
	DrawAllLines();
	InvalidateVScrollBar();

	
}

void gEditBox::RemoveText(int Start,int End)
{
	int cnlines=0,vnlines=0,tempoff=Start;
	
	while( (tempoff=Text.NextLine(tempoff))>=0) {
		if(tempoff>End) break;
		--TotalLines;
		if(CursorLine>=tempoff) ++cnlines;
		if(ViewLine>=tempoff) ++vnlines;
		
	}
	
	// UPDATE ALL POINTERS
	
	
	
	if(CursorPos>=End) {
		CursorPos-=End-Start;
		Cursory-=cnlines*Font.TextHeight();
	}
	else if(CursorPos>=Start) {
		CursorPos=Start;
		Cursory-=cnlines*Font.TextHeight();
	}
	
	if(ViewLine>=End) {
		ViewLine-=End-Start;
		viewy-=vnlines*Font.TextHeight();
	}
	else {
		if(ViewLine>Start) {
			ViewLine=Text.StartOfLine(Start);
			viewy-=vnlines*Font.TextHeight();
		}
	}

	
	Text.MemBlock::Delete(Start,End-Start);
	CursorLine=Text.StartOfLine(CursorPos);
	Cursorx=Text.TextWidth(CursorLine,CursorPos,Font);
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		viewx=Cursorx;
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
	}

	
	if(CursorLine<ViewLine) {
		viewy=Cursory;
		ViewLine=CursorLine;
	}

	if(SelStart>Start) {
		SelStart-=End-Start;
		if(SelStart<Start) SelStart=Start;
	}
	if(SelEnd>Start) {
		SelEnd-=End-Start;
		if(SelEnd<Start) SelEnd=Start;
	}
	if(SelStart==SelEnd) SelStart=SelEnd=-1;
	
	if(State&ES_LIMITEDIT) {
		if(EditMin>Start) {
			EditMin-=End-Start;
			if(EditMin<Start) EditMin=Start;
		}
		if(EditMax>Start) {
			EditMax-=End-Start;
			if(EditMax<Start) EditMax=Start;
		}
			
	}
	
}

void gEditBox::InsertText(int Start,int Bytes,gString& src)
{
	int cnlines=0,tempoff=Start,stpos=0;
	int End=Start+Bytes;

	if(State&ES_OVERWRITE) {
		int eol=Text.EndOfLine(Start);
		if(eol>Start) {
			int b=min(Bytes,eol-Start);
			Text.MemMove(Start,src,0,b);
			if(CursorPos>=Start && CursorPos<Start+b) CursorPos=Start+b;
			stpos+=b;
			Bytes-=b;
			Start+=b;
		}
	}

	if(!Text.MemBlock::Insert(Start,Bytes)) {
		// TODO: SEND AUDIBLE SIGNAL THAT OPERATION FAILED
		return;
		
	}
	Text.MemBlock::MemMove(Start,src,stpos,Bytes);
	
	TotalLines+=src.NumLines()-1;
	
	while( (tempoff=Text.NextLine(tempoff))>=0) {
		if(tempoff>End) break;
		++cnlines;
	}
	
	// UPDATE ALL POINTERS
	if(CursorPos>=Start) {
		CursorPos+=Bytes;
		Cursory+=cnlines*Font.TextHeight();
		CursorLine=CursorPos;
		CursorLine=Text.StartOfLine(CursorPos);
	}
	if(ViewLine>Start) {
		ViewLine=Text.StartOfLine(ViewLine+Bytes);
		viewy+=cnlines*Font.TextHeight();
	}
	
	Cursorx=Text.TextWidth(CursorLine,CursorPos,Font);
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		viewx=Cursorx;
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
	}
	
	// SCROLL DOWN AS NEEDED
	while(Cursory+Font.TextHeight()>viewy+viewh) {
		viewy+=Font.TextHeight();
		ViewLine=Text.NextLine(ViewLine);
	}
	if(SelStart>Start) SelStart+=Bytes;
	if(SelEnd>=Start) SelEnd+=Bytes;
	if(State&ES_LIMITEDIT) {
		if(EditMin>Start) EditMin+=Bytes;
		if(EditMax>=Start) EditMax+=Bytes;
			
	}

}

void gEditBox::InsertText(int Start,int Bytes,char *data)
{
	int cnlines=0,tempoff;
	int End=Start+Bytes;

	if(State&ES_OVERWRITE) {
		int eol=Text.EndOfLine(Start);
		if(eol>Start) {
			int b=min(Bytes,eol-Start);
			Text.MemMove(Start,data,b);
			if(CursorPos>=Start && CursorPos<Start+b) CursorPos=Start+b;

			data+=b;
			Bytes-=b;
			Start+=b;
		}
	}
	if(!Text.MemBlock::Insert(Start,Bytes,data)) {
		// TODO: SEND AUDIBLE SIGNAL THAT OPERATION FAILED
		return;
		
	}
	tempoff=Start;
	while( (tempoff=Text.NextLine(tempoff))>=0) {
		if(tempoff>End) break;
		++cnlines;
	}
	
	// UPDATE ALL POINTERS
	TotalLines=Text.NumLines();
	if(CursorPos>=Start) {
		CursorPos+=Bytes;
		Cursory+=cnlines*Font.TextHeight();
		CursorLine=Text.StartOfLine(CursorPos);
	}
	
	if(ViewLine>Start) {
		ViewLine=Text.StartOfLine(ViewLine+Bytes);
		viewy+=cnlines*Font.TextHeight();
	}
	
	Cursorx=Text.TextWidth(CursorLine,CursorPos,Font);
	if(Cursorx<viewx) {
		// SCROLL TO THE LEFT
		viewx=Cursorx;
	}
	if(Cursorx+cursorbk.width>=viewx+vieww) {
		// SCROLL TO THE RIGHT
		int shift=Cursorx+cursorbk.width-viewx-vieww;
		viewx+=shift;
	}
	// SCROLL DOWN AS NEEDED
	while(Cursory+Font.TextHeight()>viewy+viewh) {
		viewy+=Font.TextHeight();
		ViewLine=Text.NextLine(ViewLine);
	}

	if(SelStart>Start) SelStart+=Bytes;
	if(SelEnd>=Start) SelEnd+=Bytes;
	if(State&ES_LIMITEDIT) {
		if(EditMin>Start) EditMin+=Bytes;
		if(EditMax>=Start) EditMax+=Bytes;
			
	}
	

}



void gEditBox::Paste()
{
	gApplication *app=GetApplication();
	gString& st=app->Clipboard;
	InsertText(CursorPos,st.TotalUsed,st);
	ClearClipRegion();
	DrawAllLines();
	InvalidateVScrollBar();

}

// LOAD/SAVE
void gEditBox::SetText(const char *text)
{
	CursorOff();
	RemoveText(0,Text.TotalUsed-1);
	InsertText(0,strlen(text),(char *)text);
	ClearClipRegion();
	DrawAllLines();
	InvalidateVScrollBar();
	CursorOn();
}
int gEditBox::GetText(char *data,int maxbytes)
{
	int bytes=min(maxbytes,Text.Strlen());
	
	Text.MemBlock::MemMove(data,0,bytes);
	return bytes;
}


// KEYBOARD DEFAULT TABLE
// KEYBOARD MODES ARE 'L'=LOWERCASE, 'C'=UPPERCASE, 'S'=SYMBOL, 'T'=TOKEN
// FIRST WORD IS UNSHIFTED, SECOND WORD IS SHIFTED


const char gEditBoxDefaultKeys[]=
{
// L   ,  C  , 	LS  , RS  , LSH , RSH , N/A , N/A ,	

// KEY CODE 0
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the backspace key
	//#define KB_BKS  1
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   
   //! Constant for the TAN (U) key
	//#define KB_U    2
   'u'   ,  'U'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

   //! Constant for the / (Z) key
	//#define KB_DIV  3
   'z'   ,  'Z'  ,  0   ,  '/'  ,  0  ,  '/'  ,  0  ,  0  ,
   
	//! Constant for the * key
	//#define KB_MUL  4
   '*'   ,  '*'  ,  '['   ,  '\"'  ,  ']'  ,  '"'  ,  0  ,  0  ,

	//! Constant for the - key
	//#define KB_SUB  5
   '-'   ,  '-'  ,  '('   ,  '_'  ,  ')'  ,  '_'  ,  0  ,  0  ,

   //! Constant for the + key
	//#define KB_ADD  6
   '+'   ,  '+'  ,  '{'   ,  171  ,  '}'  ,  187  ,  0  ,  0  ,

	//! Constant for the ENT key
	//#define KB_ENT  7
   '\n'   ,  '\n'  ,  '&'   ,  '@'  ,  '&'  ,  '@'  ,  0  ,  0  ,

   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   
   
	//! Constant for the SYMB (P) key
	//#define KB_P    9
   'p'   ,  'P'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

   //! Constant for the COS (T) key
	//#define KB_T   10
   't'   ,  'T'  ,  0   ,  149  ,  0  , 149  ,  0  ,  0  ,

	//! Constant for the 1/X (Y) key
	//#define KB_Y   11
   'y'   ,  'Y'  ,  138   ,  '>'  ,  138  ,  '>'  ,  0  ,  0  ,

	//! Constant for the 9 key
	//#define KB_9   12
   '9'   ,  '9'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the 6 key
	//#define KB_6   13
   '6'   ,  '6'  ,  167   ,  128  ,  167  ,  176  ,  0  ,  0  ,

	//! Constant for the 3 key
	//#define KB_3   14
   '3'   ,  '3'  ,  '#'   ,  '?'  ,  '#'  , 191  ,  0  ,  0  ,

	//! Constant for the SPC key
	//#define KB_SPC 15
   ' '   ,  ' '  ,  135   ,  ','  ,  135  ,  ';'  ,  0  ,  0  ,

   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

   //! Constant for the ' (O) key
	//#define KB_O   17
   'o'   ,  'O'  ,  0   ,  '\''  ,  0  ,  157  ,  0  ,  0  ,

	//! Constant for the SIN (S) key
	//#define KB_S   18
   's'   ,  'S'  ,  0   , 152  ,  0  ,  152  ,  0  ,  0  ,

   //! Constant for the X (X) key
	//#define KB_X   19
   'x'   ,  'X'  ,  137   ,  '<'  ,  137  ,  '<'  ,  0  ,  0  ,

	//! Constant for the 8 key
	//#define KB_8   20
   '8'   ,  '8'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the 5 key
	//#define KB_5   21
   '5'   ,  '5'  ,  163   ,  '\\'  ,  163  ,  '\\'  ,  0  ,  0  ,

	//! Constant for the 2 key
	//#define KB_2   22
   '2'   ,  '2'  ,  ';'   ,  '!'  ,  ';'  ,  161  ,  0  ,  0  ,

   //! Constant for the . key
	//#define KB_DOT 23
   '.'   ,  '.'  ,  ':'   ,  '\n'  ,  ':'  ,  '\n'  ,  0  ,  0  ,

   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

   //! Constant for the EVAL (N) key
	//#define KB_N   25
   'n'   ,  'N'  ,  0   ,  150  ,  0  ,  150  ,  0  ,  0  ,

	//! Constant for the Sqrt (R) key
	//#define KB_R   26
   'r'   ,  'R'  ,  178   ,  131  ,  178  ,  131  ,  0  ,  0  ,

	//! Constant for the +/- (W) key
	//#define KB_W   27
   'w'   ,  'W'  ,  139   ,  '='  ,  139  ,  '='  ,  0  ,  0  ,
   
	//! Constant for the 7 key
	//#define KB_7   28
   '7'   ,  '7'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the 4 key
	//#define KB_4   29
   '4'   ,  '4'  ,  '$'   ,  160  ,  '$'  ,  160  ,  0  ,  0  ,

	//! Constant for the 1 key
	//#define KB_1   30
   '1'   ,  '1'  ,  '%'   ,  '~'  ,  '%'  ,  '~'  ,  0  ,  0  ,

	//! Constant for the 0 key
	//#define KB_0   31
   '0'   ,  '0'  ,  159   ,  141  ,  159  ,  141  ,  0  ,  0  ,

   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the HIST (M) key
	//#define KB_M   33
   'm'   ,  'M'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the Y^X (Q) key
	//#define KB_Q   34
   'q'   ,  'Q'  ,  0   ,  '^'  ,  0  ,  '^'  ,  0  ,  0  ,

	//! Constant for the EEX (V) key
	//#define KB_V   35
   'v'   ,  'V'  ,  0   ,  154  ,  0  ,  154  ,  0  ,  0  ,

   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,
   
   //! Constant for the F1 (A) key
	//#define KB_A   41
   'a'   ,  'A'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the F2 (B) key
	//#define KB_B   42
   'b'   ,  'B'  ,  0  ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the F3 (C) key
	//#define KB_C   43
   'c'   ,  'C'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the F4 (D) key
	//#define KB_D   44
   'd'   ,  'D'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the F5 (E) key
	//#define KB_E   45
   'e'   ,  'E'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the F6 (F) key
	//#define KB_F   46
   'f'   , 'F'   ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the APPS (G) key
	//#define KB_G   47
   'g'   ,  'G'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

   	//! Constant for the up arrow key
	//#define KB_UP  49
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the left arrow key
	//#define KB_LF  50
   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the down arrow key
	//#define KB_DN  51
	0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the right arrow key
	//#define KB_RT  52
	0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the MODE (H) key
	//#define KB_H   53
	'h'   ,  'H'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the TOOL (I) key
	//#define KB_I   54
	   'i'   ,  'I'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the VAR (J) key
	//#define KB_J   55
	   'j'   ,  'J'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the STO (K) key
	//#define KB_K   57
	   'k'   ,  'K'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the NXT (L) key
	//#define KB_L   58
	   'l'   ,  'L'  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the ALPHA key
	//#define KB_ALPHA   60
	   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the left shift key
	//#define KB_LSHIFT  61
	   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the right shift key
	//#define KB_RSHIFT  62
	   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0  ,

	//! Constant for the ON key
	//#define KB_ON  63
	   0   ,  0  ,  0   ,  0  ,  0  ,  0  ,  0  ,  0 

			
};


BOOL gEditBox::DefaultEvent(gEvent *e)
{
	gUpdate u;
	
	if(e->Type==GEVENT_KEYB) {

		// IMPLEMENT BASIC KEYBOARD HANDLING
		switch(e->Arg)
		{
		// ADD SPECIAL KEYS HERE
		case KM_PRESS | KB_LF:
		if(State&ES_RDONLY) {
			ViewLeft();
			return TRUE;
		}

		CursorOff();
		CursorLeft();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_RT:
		if(State&ES_RDONLY) {
			ViewRight();
			return TRUE;
		}
		CursorOff();
		CursorRight();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_UP:
		if(!(StyleFlags&S_MULTILINE)) return FALSE;
		if(State&ES_RDONLY) {
			ViewUp();
			return TRUE;
		}
		CursorOff();
		CursorUp();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_DN:
		if(!(StyleFlags&S_MULTILINE)) return FALSE;	
		if(State&ES_RDONLY) {
			ViewDown();
			return TRUE;
		}

		CursorOff();
		CursorDown();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_DN | SHIFT_RS | SHIFT_RSHOLD:
		if(State&ES_RDONLY) {
			ViewPageDown();
			return TRUE;
		}
		CursorOff();
		PageDown();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_UP | SHIFT_RS | SHIFT_RSHOLD:
		if(State&ES_RDONLY) {
			ViewPageUp();
			return TRUE;
		}
		CursorOff();
		PageUp();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_LF | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		//PrevWord();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_RT | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		//NextWord();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_UP | SHIFT_RS:
		if(State&ES_RDONLY) {
			ViewStartOfText();
			return TRUE;
		}
		CursorOff();
		StartOfText();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_DN | SHIFT_RS:
		if(State&ES_RDONLY) {
			ViewEndOfText();
			return TRUE;
		}

		CursorOff();
		EndOfText();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_LF | SHIFT_RS:
		if(State&ES_RDONLY) {
			ViewHome();
			return TRUE;
		}
		CursorOff();
		Home();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_RT | SHIFT_RS:
		if(State&ES_RDONLY) {
			ViewEnd();
			return TRUE;
		}

		CursorOff();
		End();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_G | SHIFT_RS:
		case KM_PRESS | KB_G | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		BeginSelection();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_H | SHIFT_RS:
		case KM_PRESS | KB_H | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		EndSelection();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_J | SHIFT_RS:
		case KM_PRESS | KB_J | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		Copy();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_K | SHIFT_RS:
		case KM_PRESS | KB_K | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		Cut();
		CursorOn();
		return TRUE;
		case KM_PRESS | KB_L | SHIFT_RS:
		case KM_PRESS | KB_L | SHIFT_RS | SHIFT_RSHOLD:
		CursorOff();
		Paste();
		CursorOn();
		return TRUE;

		
		case KM_PRESS | KB_BKS:
		{
		gEvent ep;
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_BEFORECHANGE;
		ep.Data=8;
		if(DoEventExclusive(&ep)) {
			if(!ep.Data) return TRUE;
		}	
		if(!(State&ES_RDONLY)) DelChar(TRUE);
		ep.Type=GEVENT_GUI;
		ep.Arg=GEVGUI_AFTERCHANGE;
		ep.Data=8;
		DoEventExclusive(&ep);
		return TRUE;
		}
		case KM_PRESS | KB_BKS | SHIFT_LS | SHIFT_LSHOLD:
		case KM_PRESS | KB_BKS | SHIFT_LS:
		{
			gEvent ep;
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_BEFORECHANGE;
			ep.Data=127;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}	
			if(!(State&ES_RDONLY)) DelChar(FALSE);
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_AFTERCHANGE;
			ep.Data=127;
			DoEventExclusive(&ep);
			return TRUE;
		}
		case KM_KEYDN | KB_ALPHA:
		CursorOff();
		if((InputMode&0xff)=='L') InputMode=(InputMode&(~0xff)) | 'C' | 0x300; 
		else { if((InputMode&0xff)=='C') InputMode=(InputMode&(~0xff)) | 'L' | 0x300; }
		CursorOn();
		return TRUE;

		case KM_KEYUP | KB_ALPHA:
		if(((InputMode&0x300)==0x200)) {
		CursorOff();
		if((InputMode&0xff)=='L') InputMode=(InputMode&(~0x1ff)) | 'C'; 
		else { if((InputMode&0xff)=='C') InputMode=(InputMode&(~0x1ff)) | 'L'; }
		InputMode&=~0x200;
		CursorOn();
		return TRUE;
		}
		return FALSE;
		case KM_PRESS | KB_UP | SHIFT_ALPHAHOLD:
		case KM_PRESS | KB_DN | SHIFT_ALPHAHOLD:
		// DISENGAGE ALPHA MODE BEFORE LOSING FOCUS
		if(InputMode&0x200) {
		CursorOff();
		if((InputMode&0xff)=='L') InputMode=(InputMode&(~0x1ff)) | 'C'; 
		else { if((InputMode&0xff)=='C') InputMode=(InputMode&(~0x1ff)) | 'L'; }
		InputMode&=~0x200;
		CursorOn();
		}
		return FALSE; 	// ALLOW DEFAULT POST-PROCESSING
		
		case KM_PRESS | KB_ENT :
		// DO NOT REARRANGE!! - MUST FALL TO default:
		{
			gEvent ep;
			ep.Type=GEVENT_GUI;
			ep.Arg=GEVGUI_ENTER;
			ep.Data=0;
			if(DoEventExclusive(&ep)) {
				if(!ep.Data) return TRUE;
			}
		}
		if(!(StyleFlags&S_MULTILINE)) return FALSE;	
		
		// USE DEFAULT KEY BEHAVIOR
		default:
			if(KM_MESSAGE(e->Arg)==KM_PRESS) {
				int key,shift;
				char a;
				key=KM_KEY(e->Arg);
				shift=KM_SHIFTPLANE(e->Arg);
				switch(shift) {
				case SHIFT_LS | SHIFT_LSHOLD:
				shift=4;
				break;
				case SHIFT_RS | SHIFT_RSHOLD:
				shift=5;
				break;
				case SHIFT_LS:
				shift=2;
				break;
				case SHIFT_RS:
				shift=3;
				break;
				case SHIFT_ALPHAHOLD:
				if((InputMode&0xff)=='L') { shift=0; InputMode=InputMode&(~0x100); } 
				else if((InputMode&0xff)=='C') { shift=1; InputMode=InputMode&(~0x100); }
				break;
				default:
				if((InputMode&0xff)=='L') shift=0;
				if((InputMode&0xff)=='C') shift=1;
				break;
				}
			a=gEditBoxDefaultKeys[(key<<3)+shift];
			if(a)  {
				gEvent ep;
				ep.Type=GEVENT_GUI;
				ep.Arg=GEVGUI_BEFORECHANGE;
				ep.Data=a;
				if(DoEventExclusive(&ep)) {
					if(!ep.Data) return TRUE;
				}	
				if(!(State&ES_RDONLY)) InsertChar((int)a);
				ep.Type=GEVENT_GUI;
				ep.Arg=GEVGUI_AFTERCHANGE;
				ep.Data=a;
				DoEventExclusive(&ep);
				return TRUE;
				}
			}
		return FALSE;

		}

	}
	
	if(e->Type==GEVENT_GUI) {
		switch(e->Arg)
		{
		case GEVGUI_GETFOCUS:
			StyleFlags|=S_FOCUSED;
			u.clipx=ncx;
			u.clipy=ncy;
			u.clipx2=ncx+ncwidth-1;
			u.clipy2=ncy+ncheight-1;
			Parent->Invalidate(&u,this);
			StartCursorBlink();
			return TRUE;
		case GEVGUI_LOSEFOCUS:
			StopCursorBlink();
			StyleFlags&=~S_FOCUSED;
			u.clipx=ncx;
			u.clipy=ncy;
			u.clipx2=ncx+ncwidth-1;
			u.clipy2=ncy+ncheight-1;
			Parent->Invalidate(&u,this);
			return TRUE;
		}
		
	}
	if(e->Type==GEVENT_IDLE) {
		if((StyleFlags&S_FOCUSED) && (State&ES_CURSORBLINK)) {
				tmr_t now=tmr_ticks();
				if(now-lastblink>=CURSORBLINKSPEED) {
					if(State&ES_SHOWCURSOR) { CursorOff(); State&=~ES_SHOWCURSOR; }
					else { CursorOn(); State|=ES_SHOWCURSOR; }
					if(now-lastblink>(CURSORBLINKSPEED<<4)) lastblink=now;
					else lastblink+=CURSORBLINKSPEED;
					return TRUE;

				}
			}
	}

	return FALSE;

	
}


void gEditBox::DrawVScrollBar()
{
	gControl *P=Parent;

	int x,y,clen,cpos,h;

	x=ncx+ncwidth-SCROLLBARWIDTH;
	y=ncy;
	h=ncheight;
	if(StyleFlags&S_HASBORDER) { ++y; --x; h-=2*BORDERWIDTH; }
	
	// DRAW FRAME
	P->VLine(x,y,y+h-1,GGRAY7);
	P->Rect(x+1,y,x+SCROLLBARWIDTH-2,y+h-1,GGRAY3);
	P->HLine(y,x+1,x+SCROLLBARWIDTH-1,GGRAY7);
	P->VLine(x+SCROLLBARWIDTH-1,y,y+h-1,GWHITE);
	P->HLine(y+h-1,x+1,x+SCROLLBARWIDTH-1,GWHITE);

	// DRAW CURSOR
	if(TotalLines>0) clen=((h-3)*(DispLines))/TotalLines;
	else return;
	if(clen<1) clen=1;
	if(clen>h-3) clen=h-3;
	cpos=(viewy*(h-3-clen))/((TotalLines-DispLines)*Font.TextHeight());

	P->Rect(x+1,y+cpos+1,x+SCROLLBARWIDTH-2,y+cpos+clen+1,GBLACK);
	
}

void gEditBox::InvalidateVScrollBar()
{
	int x,y,h;

	x=ncx+ncwidth-SCROLLBARWIDTH;
	y=ncy;
	h=ncheight;
	if(StyleFlags&S_HASBORDER) { ++y; --x; h-=2*BORDERWIDTH; }
	
	gUpdate u;
	u.clipx=x;
	u.clipx2=x+SCROLLBARWIDTH-1;
	u.clipy=y;
	u.clipy2=y+h-1;
	Parent->Invalidate(&u,this);
}

gEditBox::gEditBox(gControl& _Parent,int x, int y,int width,int height,gStyles _Style, gFont *f)
: gControl(_Parent)
{
	int w,h;

	memcpy(&Font,f,sizeof(gFont));
	
	if(width<0) {
		// WIDTH SPECIFIED IN CHARACTERS
		w=-width*Font.MaxWidth;
		if(_Style&S_VSCROLLBAR) w+=SCROLLBARWIDTH;
		if(_Style&S_HASBORDER) w+=2*BORDERWIDTH+2;
	}
	else w=width;

	if(height<0) {
		// HEIGHT SPECIFIED IN LINES OF TEXT
		h=-height*Font.TextHeight();
		//if(_Style&S_HSCROLLBAR) h+=SCROLLBARWIDTH;
		if(_Style&S_HASBORDER) h+=2*BORDERWIDTH+2;
	}
	else {
		int tmph=Font.TextHeight();
		if(_Style&S_HASBORDER) tmph+=2*BORDERWIDTH+2;
		if(tmph>height) h=tmph;
		else h=height;
	}
	
	StyleFlags=_Style;

	ncx=x;
	ncy=y;
	posx=x;
	posy=y;
	ncwidth=w;
	ncheight=h;

	// CALCULATE CLIENT AREA HEIGHT BASED ON STYLES
	
	if(StyleFlags & S_HASBORDER) { posx+=BORDERWIDTH+1; posy+=BORDERWIDTH+1; h-=2*BORDERWIDTH+2; w-=2*BORDERWIDTH+2; }
//	if(StyleFlags&S_HSCROLLBAR) {
//		h-=SCROLLBARWIDTH;
//	}
	if(StyleFlags&S_VSCROLLBAR) {
		w-=SCROLLBARWIDTH;
	}
	
	DispLines=h/Font.TextHeight();
	if(DispLines<=0) DispLines=1;

	vieww=w;
	viewh=DispLines*Font.TextHeight();
	viewx=viewy=0;
	if(!Resize(w,h)) InsufMemoryError();

	ClearClipRegion();
	char a=0;
	Text.Append(1,&a);
	TotalLines=1;
	
	Color=GBLACK;
	BkColor=GWHITE;
	SelColor=GGRAY5;
	CursorPos=CursorLine=0;
	SelStart=SelEnd=-1;
	State=0x10000;	// START WITH CURSOR OFF
	Cursorx=Cursory=0;
	ViewLine=0;
	InputMode='L';
	EditMin=-1;
	EditMax=-1;
	
	cursorbk.x=cursorbk.y=0;
	cursorbk.width=Font.MaxWidth;
	cursorbk.addr=(int *)new char [ (cursorbk.width*Font.TextHeight()+1)>>1];
	if(!cursorbk.addr) InsufMemoryError();
	
}
gEditBox::~gEditBox()
{
}

void gEditBox::SetEditLimits(int StartPos,int EndPos)
{
	if(!Text.IsValidOffset(StartPos)) StartPos=0;
	if(!Text.IsValidOffset(EndPos)) EndPos=Text.EndOfText();
	
	EditMin=StartPos;
	EditMax=EndPos;
	State|=ES_LIMITEDIT;
	if(CursorPos<EditMin || CursorPos>=EditMax) MoveCursor(EditMin);
	return;

}

void gEditBox::GetEditLimits(int *StartPos,int *EndPos)
{
	*StartPos=EditMin;
	*EndPos=EditMax;
	return;
}

void gEditBox::ClearEditLimits()
{
	EditMin=-1;
	EditMax=-1;
	State&=~ES_LIMITEDIT;
}

void gEditBox::SetReadOnly(BOOL rdonly)
{
	if(rdonly) {
		if(State&ES_RDONLY) return;
		State|=ES_RDONLY;
		StopCursorBlink();
		CursorOff();
	}
	else {
		if(!(State&ES_RDONLY)) return;
		State&=~ES_RDONLY;
		MoveCursor(CursorPos);
		StartCursorBlink();
		CursorOn();
	}
}

BOOL gEditBox::IsReadOnly()
{
	return (State&ES_RDONLY)? TRUE:FALSE;
}

int gEditBox::GetText(char *data,int maxbytes,int StartPos,int EndPos)
{
	int bytes=min(maxbytes,EndPos-StartPos+1);
	
	Text.MemBlock::MemMove(data,StartPos,bytes);
	return bytes;
}

void gEditBox::SetOverwriteMode(BOOL ovrwrite)
{
	if(ovrwrite) State|=ES_OVERWRITE;
	else State&=~ES_OVERWRITE;
}

void gEditBox::LimitTextSize(int limit,int trimwhere)
{
	int currentsize=Text.Strlen();
	if(currentsize<=limit) return;
	
	if(trimwhere==TRIM_BEGIN) {
		int pos,prev=Text.StartOfText();
		int minpos=currentsize-limit;
		int count=0;
		do {
			pos=Text.NextLine(prev);
			if(pos<0) { pos=prev; break; }
			prev=pos;
			++count;
		} while(pos<minpos);
		
		if(pos==Text.StartOfText()) return;
		CursorOff();
		
		RemoveText(0,pos);

		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		CursorOn();

		return;
		
	}
	
	if(trimwhere==TRIM_END) {
		int prev=Text.StartOfLine(limit);
		
		CursorOff();
		RemoveText(prev,Text.EndOfText());

		ClearClipRegion();
		DrawAllLines();
		InvalidateVScrollBar();
		CursorOn();
		return;
		
	}
	
}



