//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***


/*    qlog10.c "Quick" log10
    20050123 Daniel Herring
    Provide a "quick" >low-memory-footprint< method for base-10 logarithms.
    Taylored for the HP series calculators which only have exponents up
    to +/- 500.

    Uses at most 30 multiplies (over the range 10^(+/-1000)).

    There are numerically better methods, but this was easy to implement.

    Believed to be correct, but should be double-checked.
*/


/**
 * find y and return so that x=y*(10**return)
 *
 * @param x source number
 * @param *y location for the remainder term
 * @return integer log10 of x
 */
 
#include <sys.h>
#include <math.h>

// ROMLIB CHECKED
// < ibl ; 2006-04-03 >


int qlog10(double x, double *y)
{
    int power=0;
    
    if ( x <= 0.0 )
        sys_error("qlog10: Invalid Argument",-1);
    
    if(x>1)
    {
        while(x>1E100)
        {
            x*=1E-100;
            power+=100;
        }
        while(x>1E10)
        {
            x*=1E-10;
            power+=10;
        }
        while(x>1)
        {
            x*=0.1;
            power+=1;
        }
    }
    else if(x<1)
    {
        while(x<1E-100)
        {
            x*=1E100;
            power-=100;
        }
        while(x<1E-10)
        {
            x*=1E10;
            power-=10;
        }
        while(x<1)
        {
            x*=10;
            power-=1;
        }
    }

    *y=x;
    return power;
}

