//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: launch_fixsto.c 393 2007-03-04 23:17:29Z ingo $

#include <hpgcc3.h>
#include "launcher.h"


// ALIGN CODE OBJECT PRIOR TO BEING STORED IN FLASH

void __launch_fixobj(int flash_addr,int ram_addr,SAT_CPU *cpu)
{
int nnibbles;
int slack,startcode,offset;

// GET POINTER TO ARM CODE


if(__launch_peek(ram_addr,5,cpu)!=SAT_DOARMC) return;	// NO ARM CODE OBJECT, NOTHING TO DO
if(__launch_peek(ram_addr+10,5,cpu)!=0x34343) return; // INVALID EXEC, DO NOTHING
slack=__launch_peek(ram_addr+24,2,cpu);
if(!slack) return;										// NO SLACK SPACE, DO NOTHING

nnibbles=__launch_peek(ram_addr+5,5,cpu)-43 - (2<<slack);	// GET CODE SIZE

startcode=__launch_peek(ram_addr+26,6,cpu)+ram_addr+48;

offset=flash_addr+48;		// START OF SLACK SPACE
offset+=(2<<slack)-1;
offset&=~2047;	// TINY-PAGE ALIGN
offset-=flash_addr+48;

if(offset<0) return;				// RETURN IF SLACK SPACE IS INSUFFICIENT TO ALIGN
if(offset>=(2<<slack)) return;

__launch_memmove(ram_addr+48+offset,startcode,nnibbles,cpu);

__launch_poke(ram_addr+26,offset,6,cpu);		// STORE NEW OFFSET

// DONE, RETURN
}


unsigned int inline __launch_addCRC(unsigned int crc,unsigned int nibble)
{
   return ((((crc^nibble) & 15) * 4225)) ^(crc>>4);
}



// PATCH LIBRARY

void __launch_fixlib(int flash_addr,int ram_addr,SAT_CPU *cpu)
{
int crc_addr,namelen, nptrs;
int addr,obj_addr;
unsigned int crc;
// GET POINTER TO ARM CODE

if(__launch_peek(ram_addr,5,cpu)!=SAT_DOLIB) return;	// NO LIBRARY OBJECT, NOTHING TO DO


// SCAN LIBRARY OBJECT FOR ARMCODE OBJECTS


crc_addr=ram_addr+1+__launch_peek(ram_addr+5,5,cpu);		// GET CRC ADDRESS

addr=ram_addr+10;

namelen=__launch_peek(addr,2,cpu);
addr+=2;

if(namelen) addr+=(namelen+1)<<1;

addr+=13;

addr+=__launch_peek(addr,5,cpu);	// POINT TO HASH TABLE
nptrs=__launch_peek(addr+5,5,cpu)-5;
addr+=10;

while(nptrs>0)
{
obj_addr=(addr+__launch_peek(addr,5,cpu))&0xfffff;
__launch_fixobj(flash_addr+(obj_addr-ram_addr),obj_addr,cpu);
addr+=5;
nptrs-=5;
}

// RECALCULATE LIBRARY CRC
crc=0;
addr=ram_addr+5;
while(addr<crc_addr) {
crc=__launch_addCRC(crc,__launch_peek(addr,1,cpu));
++addr;
}

__launch_poke(crc_addr,crc,4,cpu);		// STORE CORRECTED CRC


return;

}




// ON ENTRY: D1+8 NIBBLES = FLASH ADDR OF PROLOG OF OBJECT TO STORE
//			 R1.A         = RAM ADDR OF PROLOG

int launch_fixsto(int pc,SAT_CPU *cpu)
{
int ram_addr,flash_addr;

// GET POINTER TO ARM CODE

flash_addr=(cpu->D1&0xfffff)+8;
ram_addr=cpu->R1&0xfffff;

__launch_fixlib(flash_addr,ram_addr,cpu);

return pc;

}

// ON ENTRY: R2.A = FLASH ADDR OF FIRST SLOT
// 			 R1.A = TOTAL NIBLES TO LAST SLOT
//			 R0.A = RAM ADDR OF FIRST SLOT

int launch_packbank(int pc,SAT_CPU *cpu)
{
int ram_addr,flash_addr,end_addr;
int slotsize;
// GET POINTER TO ARM CODE

flash_addr=cpu->R2&0xfffff;
ram_addr=cpu->R0&0xfffff;

end_addr=ram_addr+(cpu->R1&0xfffff);

while(ram_addr<end_addr) {
__launch_fixlib(flash_addr+8,ram_addr+8,cpu);
slotsize=__launch_peek(ram_addr+2,5,cpu)+2;
ram_addr+=slotsize;
flash_addr+=slotsize;
}

return pc;

}



