//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***



// $Id: legacy_launch.c 533 2008-01-25 23:55:41Z claudio $

#include <hpgcc3.h>
#include "launcher.h"



#define STANDARD_ROMP_EVAL (pc+7)


typedef int (*ENTRY_POINT)(unsigned int,unsigned int);


int __launch_checkromlib(unsigned int rlibver);

// SCAN ARM OBJECT POINTED TO BY A.A AND EVAL IT

int launch_legacy(int pc,SAT_CPU *cpu)
{
register int entry_pt;
struct {
int obj_addr,code_addr,type,size,prolog,ram_req,kos_mem;
int sat_backupB,mmubackup[2];
int *mmu_table;
} str;
// GET POINTER TO ARM CODE


str.obj_addr=cpu->A&0xfffff;


str.type=__launch_peek(str.obj_addr,5,cpu);

if(str.type==SAT_DOROMP) {

// ATTEMPT TO EVAL A ROMPOINTER

str.obj_addr=launch_rcl_romp(str.obj_addr,cpu);

if(!str.obj_addr) {
// UNABLE TO FIND ROMP, RESUME STANDARD ROMP EVAL
return STANDARD_ROMP_EVAL;
}

str.type=__launch_peek(str.obj_addr,5,cpu);

// ROMPTR DOES NOT CONTAIN AN ARMCODE OBJECT, DO STANDARD EVAL
if(str.type!=SAT_DOARMC) return STANDARD_ROMP_EVAL;

}



str.size=__launch_peek(str.obj_addr+5,5,cpu);
str.prolog=__launch_peek(str.obj_addr+10,5,cpu);



if(str.type==SAT_DOCSTR || str.type==SAT_DOARMC) {

if(str.prolog==0x34343) {
// NEW EXECUTABLE FORMAT
str.code_addr=__launch_peek(str.obj_addr+26,6,cpu);
str.code_addr+=str.obj_addr+48;
// CHECK FOR ROMLIB VERSION NUMBER
if(!__launch_checkromlib(__launch_peek(str.obj_addr+16,8,cpu))) return pc;




entry_pt=__launch_peek(str.obj_addr+40,8,cpu);
str.ram_req=__launch_peek(str.obj_addr+32,8,cpu);
str.obj_addr+=str.size+5;	// POINT AT END-OF-CODE
}
else {
// DO A LEGACY LAUNCH
if(str.prolog==0xb34c) { // NON-VC EXEC, TREAT AS A MISALIGNED VC EXEC
str.code_addr=__launch_peek(str.obj_addr+15,1,cpu);
str.code_addr+=str.obj_addr+16;
}
else {
if(str.prolog==0xa94c) { // VC EXECUTABLE FORMAT
str.code_addr=__launch_peek(str.obj_addr+15,5,cpu);
str.code_addr+=str.obj_addr+20;
}
else {
return pc;	// DO-NOTHING, NOT AN ARM PROGRAM
}
}
str.obj_addr+=str.size-3;	// POINT AT LINKER STRUCTURE
if(__launch_peek(str.obj_addr,8,cpu)==0x3176B34C) {
// GET ENTRY POINT
entry_pt=0x01220000+__launch_peek(str.obj_addr-16,8,cpu)-__launch_peek(str.obj_addr-8,8,cpu);
// GET MEMORY REQUIREMENTS
str.ram_req=__launch_peek(str.obj_addr-24,8,cpu);
str.obj_addr-=32;
str.obj_addr-=__launch_peek(str.obj_addr,8,cpu)<<4;
}
else {
str.ram_req=0X4A800400;		// DEFAULT VC EXECUTION W/FULL RAM
entry_pt=0x01220000;
str.obj_addr+=8;
}

// CONVERT LEGACY MEMORY REQ. TO NEW FORMAT
str.ram_req=((str.ram_req+1023)>>10&0xff)|((str.ram_req>>5)&0x40000)|(((~str.ram_req)>>13)&0x10000)|(((~str.ram_req)>>11)&0x20000)|0x100;

}
// DO FINAL LAUNCH

// ALLOCATE RAM AS REQUESTED
str.mmu_table=launch_alloc_ram(str.ram_req,&(str.kos_mem),cpu);

if(!str.mmu_table) return pc;		// RETURN IF NOT ENOUGH MEMORY TO CREATE MMU_TABLE

// UPDATE MMU TABLE W/CODE, TAKES RAM FROM MMU TO REALIGN CODE IF NEEDED
if(!launch_align_code(str.mmu_table,(__launch_s2a(str.code_addr,cpu)<<1)+(str.code_addr&1),str.obj_addr-str.code_addr,cpu))
{
// NOT ENOUGH MEMORY TO RUN
// DO-NOTHING AND RETURN;
// RETURN ALLOCATED MEMORY AS NEEDED
if(str.kos_mem) kos_free((void *)str.kos_mem);

return pc;

}

// PRESERVE SATURN REGISTER MODIFIED DURING LAUNCH
str.sat_backupB=cpu->B;
cpu->B=0;
str.mmu_table= (int *) (((unsigned int)str.mmu_table)&(~0xfff));
while(*str.mmu_table) { str.mmu_table++; ++cpu->B; }	// COUNT NUMBER OF ENTRIES IN MMU TABLE

// mmu_table POINTS TO END OF MMU TABLE
// ********** DEBUG ONLY **************
/*
cpu->R1=0x33333;
cpu->R1h=str.ram_req;
cpu->R2=entry_pt;
cpu->R2h=str.kos_mem;
cpu->R3=str.mmu_table;
if(str.kos_mem) kos_free((void *)str.kos_mem);
return pc;
*/
// ********** DEBUG ONLY **************


//****************************************************
// FINAL LAUNCH STAGE
//****************************************************



register int *stackptr asm("sp");


// INSTALL VIRTUAL MEMORY

// PRESERVE ORIGINAL VIRTUAL MEMORY
str.mmubackup[0]=*((unsigned int *)0x07f00044);
str.mmubackup[1]=*((unsigned int *)0x07f00048);

// FLUSH ALL CACHES BEFORE SWITCHING MEMORY BLOCKS
__launch_flush_cache();


*((unsigned int *)0x07f00044) = (((unsigned int)str.mmu_table)&(~0xfff)) | 0x11;	// DATA - COARSE PAGES
*((unsigned int *)0x07f00048) = (((unsigned int)str.mmu_table)&(~0xfff)) | 0x13;  // CODE - FINE PAGES



str.type=(int)stackptr;
// CALCULATE NEW STACK POINTER
register volatile unsigned int *stacktemp=(unsigned int *)(((((unsigned int)str.mmu_table)&0xfff)<<10)+0x01100000-4);
// PRESERVE OLD STACK ADDRESS
*stacktemp=str.type;

// MOVE STACK TO VIRTUAL ADDRESS

// INVALIDATE ALL MEMORY REFERENCES
asm volatile ("" : : : "memory");
// FORCE GCC TO WRITE TO THE SP REGISTER, OTHERWISE IT IGNORES stackptr=...;
asm volatile ("mov %0,%1" : "=r" (stackptr) : "r" (stacktemp) );

// FINAL JUMP
pc=(*((ENTRY_POINT)entry_pt))((unsigned int)pc,(unsigned int)cpu);



// RESTORE OLD STACK ADDRESS
asm volatile ("ldr %0,%1" : "=r" (stackptr) : "g" (*stackptr));
asm volatile ("" : : : "memory");

// FLUSH ALL CACHES
__launch_flush_cache();

// UNINSTALL VIRTUAL MEMORY
*((unsigned int *)0x07f00044) = str.mmubackup[0];	// DATA - COARSE PAGES
*((unsigned int *)0x07f00048) = str.mmubackup[1];  // CODE - FINE PAGES

// RETURN ALLOCATED MEMORY AS NEEDED
if(str.kos_mem) kos_free((void *)str.kos_mem);

// RESTORE AFFECTED SATURN REGISTERS
cpu->B=str.sat_backupB;


// DONE!
}

// DO-NOTHING AND RETURN
return pc;
}



int launch_romlibver()
{
return MAKE_ROMLIB_VER(ROMLIB_BUILD);
}


int __launch_checkromlib(unsigned int rlibver)
{
unsigned int f;
int rlibnum = launch_romlibver();

for(f=0;f<4;++f)
{
if( (rlibnum&0xff)< (rlibver&0xff)) return 0;
rlibver>>=8;
rlibnum>>=8;
}

return 1;
}

